/**
 * This code accompanies the paper:
 *
 * "A Simple and Effective Relevance-Based Point Sampling for 3D Shapes"
 * E. Rodola, A. Albarelli, D. Cremers, A. Torsello
 * Pattern Recognition Letters, 59(1), pp 41-47, 2015
 *
 * Please cite the paper above if you use this code.
 *
 * Written by Emanuele Rodola.
 **/
#ifndef MESH_H
#define MESH_H

#include <vector>
#include <list>

template <typename T>
struct vec3d
{
  T x,y,z;

  explicit vec3d() : x(0), y(0), z(0) {}
  vec3d(T x_, T y_, T z_) : x(x_), y(y_), z(z_) {}

  template <typename S>
  vec3d(const vec3d<S>& s) : x(T(s.x)), y(T(s.y)), z(T(s.z)) {}
  
  vec3d<T> operator+(const vec3d<T>& p) const
  {
	 return vec3d<T>(x + p.x, y + p.y, z + p.z);
  }

  vec3d<T> operator-(const vec3d<T>& p) const
  {
	 return vec3d<T>(x - p.x, y - p.y, z - p.z);
  }

  vec3d<T> operator-() const
  {
	 return vec3d<T>(-x, -y, -z);
  }
  
  vec3d<T>& operator+=(const vec3d<T>& p)
  {
	 x += p.x;
	 y += p.y;
	 z += p.z;
	 return *this;
  }
  
  template <typename Scalar>
  vec3d<T>& operator/=(const Scalar& s) {
	 const T i = T(1) / T(s);
	 x *= i;
	 y *= i;
	 z *= i;
	 return *this;
  }
};

// ||p1 - p2||^2
template <typename T>
inline double squared_dist(const vec3d<T>& p1, const vec3d<T>& p2)
{
  T x = p1.x - p2.x;
  T y = p1.y - p2.y;
  T z = p1.z - p2.z;
  return ((x*x) + (y*y) + (z*z));
}

template <typename T>
inline T dot_product(const vec3d<T>& v1, const vec3d<T>& v2) 
{
	return (v1.x*v2.x) + (v1.y*v2.y) + (v1.z*v2.z);
}

template <typename T>
inline double magnitude(const vec3d<T>& v1)
{
	return std::sqrt((v1.x*v1.x) + (v1.y*v1.y) + (v1.z*v1.z));
}

template <typename T>
inline double normalize(vec3d<T>& p)
{
  const double n = magnitude(p);
  if (n==0.) {
	 p.x = 0;
	 p.y = 0;
	 p.z = 0;
  }
  else {
	 p.x /= n;
	 p.y /= n;
	 p.z /= n;
  }
  return n;
}

template <typename T>
inline vec3d<T> cross_product(const vec3d<T>& v1, const vec3d<T>& v2)
{
  return vec3d<T>(
		(v1.y*v2.z) - (v1.z*v2.y),
		(v1.z*v2.x) - (v1.x*v2.z),
		(v1.x*v2.y) - (v1.y*v2.x)
		);
}

template <typename Iterator>
inline double median(Iterator start, Iterator end)
{
  const typename Iterator::difference_type n = end - start;
  if (n <= 0) return 0.;
  if (n % 2 == 0)
	 return (*(start+(n/2)) + *(start+(n/2-1))) / 2.;
  else
	 return *(start + ( (n-1) / 2 ));
}

class mesh_t
{
public:
	
   struct vertex_data {
      vertex_data() : n_tris(0) {}
      vec3d<double> p;
      size_t n_tris; // no. of triangles this vertex belongs to
      std::list<int> tri_indices; // triangles this vertex belongs to
   };

   struct triangle_data {
      int p0, p1, p2; // indices of the 3 vertices
	  vec3d<double> n;
   };
   
   std::vector<vertex_data> vertices;
   std::vector<triangle_data> triangles;
   
public:
   
   void mesh_t::put_vertices(const std::vector< vec3d<double> >& points)
	{
	   const size_t n_points = points.size();
	   vertices.resize(n_points);

	   // Put the points in the appropriate structure for the kd-tree
	   // and in the vertices collection for future indexed reference

	   for (size_t k=0; k<n_points; ++k)
	   {
		  // indexed vertices
		  vertex_data v;
		  v.p = vec3d<double>(points[k]);
		  v.n_tris = 0;
		  vertices[k] = v;
	   }
	}
	
	void mesh_t::add_triangle(int p0, int p1, int p2)
	{
	   triangle_data td;

	   td.p0 = p0;
	   td.p1 = p1;
	   td.p2 = p2;
	   
	   const vec3d<double>& p3d0 = vertices.at(p0).p;
	   const vec3d<double>& p3d1 = vertices.at(p1).p;
       const vec3d<double>& p3d2 = vertices.at(p2).p;

	   td.n = vec3d<double>( cross_product(p3d1-p3d0, p3d2-p3d0) );
       normalize(td.n);
	   
	   triangles.push_back(td);
	   const int tri_idx = static_cast<int>(triangles.size() - 1);

	   vertices.at(p0).tri_indices.push_back(tri_idx);
	   vertices.at(p0).n_tris++;
	   vertices.at(p1).tri_indices.push_back(tri_idx);
	   vertices.at(p1).n_tris++;
	   vertices.at(p2).tri_indices.push_back(tri_idx);
	   vertices.at(p2).n_tris++;
	}
	
	struct seed_t
	{
	   int v_idx;
	   int hops;
	   seed_t(int i, int n) : v_idx(i), hops(n) {}
	};
	
	/**
	 * Get the immediate neighbors of a given vertex in all the rings up to the desired level.
	 * The given vertex is not included.
	 *
	 * @param v      Vertex index.
	 * @param steps  Number of rings. Must be >= 1.
	 * @param neighs Indices to the vertices in the rings, in no specific order.
	 * @return The final number of neighbors is returned.
	 * @sa get_boundaries() and get_vertex_imm_neighbors()
	 */
	size_t mesh_t::get_vertex_ring(size_t v, int steps, std::vector<size_t>& neighs) const
	{
	   if (steps < 1)
		  throw std::invalid_argument("mesh::get_vertex_ring(): steps must be >= 1.");

	   std::vector<bool> is_ring(vertices.size(), false);

	   std::queue<seed_t> seeds;
	   seeds.push( seed_t(v,0) );

	   while(!seeds.empty())
	   {
		  const seed_t s = seeds.front(); seeds.pop();
		  is_ring[s.v_idx] = true;

		  if (s.hops < steps)
		  {
			 for (std::list<int>::const_iterator it=vertices.at(s.v_idx).tri_indices.begin(); it!=vertices.at(s.v_idx).tri_indices.end(); ++it)
			 {
				const int p0 = triangles.at(*it).p0;
				const int p1 = triangles.at(*it).p1;
				const int p2 = triangles.at(*it).p2;
				if (!is_ring[p0]) { seeds.push( seed_t(p0,s.hops+1) ); is_ring[p0] = true; }
				if (!is_ring[p1]) { seeds.push( seed_t(p1,s.hops+1) ); is_ring[p1] = true; }
				if (!is_ring[p2]) { seeds.push( seed_t(p2,s.hops+1) ); is_ring[p2] = true; }
			 }
		  }
	   }

	   neighs.clear();
	   for (size_t i=0; i<is_ring.size(); ++i)
		  if (is_ring[i] && i != v) neighs.push_back(i);

	   return neighs.size();
	}
	
	/**
	 *
	 */
	double mesh_t::get_resolution() const
	{
	  const size_t ntris = triangles.size();

	  if (ntris == 0)
		return 0.;

	  // this is inefficient and inaccurate...

	  std::vector<double> lengths;

	  for (size_t k=0; k<ntris; ++k) // consider all the edges of all the triangles
	  {
		 const vec3d<double>& p0 = vertices[ triangles[k].p0 ].p;
		 const vec3d<double>& p1 = vertices[ triangles[k].p1 ].p;
		 const vec3d<double>& p2 = vertices[ triangles[k].p2 ].p;

		 lengths.push_back( std::sqrt(squared_dist(p0,p1)) );
		 lengths.push_back( std::sqrt(squared_dist(p2,p1)) );
		 lengths.push_back( std::sqrt(squared_dist(p0,p2)) );

	  } // next triangle

	  std::sort(lengths.begin(),lengths.end());
	  //lengths.resize( std::unique(lengths.begin(),lengths.end()) - lengths.begin() );

	  double res;
	  if (lengths.empty())
		 res = 0.;
	  else
		 res = median(lengths.begin(), lengths.end());

	   return res;
	}
};
   
#endif
